﻿/*===============================================================================
Copyright (C) 2022 Immersal - Part of Hexagon. All Rights Reserved.

This file is part of the Immersal SDK.

The Immersal SDK cannot be copied, distributed, or made available to
third-parties for commercial purposes without written permission of Immersal Ltd.

Contact sdk@immersal.com for licensing requests.
===============================================================================*/

using UnityEngine;
using System;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using Immersal.REST;
using Unity.Collections.LowLevel.Unsafe;
using SC.XR.Unity;
using System.Collections;
using UnityEngine.UI;

namespace Immersal.AR
{
    public class XRLocalizer : LocalizerBase
	{
		public RawImage img;

		private static XRLocalizer instance = null;

		Vector3 offset = new Vector3(0.03f, -0.01f, 0);
		[HideInInspector]
		public WebCamTexture webTex;
		[HideInInspector]
		public string deviceName;
		private static MeshRenderer mat;
		float _width;
		float _height;

		public static XRLocalizer Instance
		{
			get
			{
#if UNITY_EDITOR
				if (instance == null && !Application.isPlaying)
				{
					instance = UnityEngine.Object.FindObjectOfType<XRLocalizer>();
				}
#endif
				if (instance == null)
				{
					Debug.LogError("No NRLocalizer instance found. Ensure one exists in the scene.");
				}
				return instance;
			}
		}

		void Awake()
		{
			if (instance == null)
			{
				instance = this;
			}
			if (instance != this)
			{
				Debug.LogError("There must be only one NRLocalizer object in a scene.");
				UnityEngine.Object.DestroyImmediate(this);
				return;
			}
		}

        public override void Start()
        {
			base.Start();
			m_Sdk.RegisterLocalizer(instance);
			StartCoroutine(CallCamera());
			//Invoke("startTrack",1f);
		}

		public void startTrack()
        {
			Localize();

			Invoke("startTrack", 1f);
		}
        public override void StartLocalizing()
        {
          //  CamTexture.Play();
            base.StartLocalizing();
        }

		public override void StopLocalizing()
		{
			base.StopLocalizing();
		//	CamTexture.Stop();
		}

        public override void Pause()
        {
            base.Pause();
		//	CamTexture.Pause();
        }

		float tcount;
		protected override void Update()
		{
			//   isTracking = NRFrame.SessionStatus == SessionState.Running;
			isTracking = true;

			base.Update();
			tcount += tcount + Time.deltaTime;
			if(tcount>0.1f)
            {
				StartCoroutine(RenderTexturesScreenCapture());
				tcount = 0;
			}
		}

        public override void OnDestroy()
        {
           // CamTexture.Stop();


			base.OnDestroy();
        }


		IEnumerator CallCamera()
		{
			yield return null;

				WebCamDevice[] devices = WebCamTexture.devices;
				deviceName = devices[0].name;
				if (webTex != null)
				{
					webTex.Stop();
					webTex = null;
				}
				//Set RGBCamera rect   
				webTex = new WebCamTexture(deviceName, 1280, 720, 30);//
				img.texture = webTex;
				webTex.Play();//Start
		}

		public void StopCamera()
		{
				webTex.Stop();
				webTex = null;
				StopCoroutine(CallCamera());
		}
		public override async void Localize()
		{

			if (m_PixelBuffer != IntPtr.Zero)
			{
				stats.localizationAttemptCount++;

				Vector4 intrinsics;
				//Pose headPose = new Pose(GSXRManager.Instance.head.position, GSXRManager.Instance.head.rotation);// NRFrame.HeadPose;
				//Pose rgbCameraFromEyePose = new Pose(GSXRManager.Instance.head.position, GSXRManager.Instance.head.rotation);// NRFrame.EyePoseFromHead.RGBEyePose;
				Vector3 camPos = OpenXRCamera.Instance.head.position;
				Quaternion camRot = OpenXRCamera.Instance.head.rotation;
				int width = webTex.width;
				int height = webTex.height;
				Vector3 pos = Vector3.zero;
				Quaternion rot = Quaternion.identity;
				GetIntrinsics(out intrinsics);

				float startTime = Time.realtimeSinceStartup;

				Task<int> t = Task.Run(() =>
				{
					return Immersal.Core.LocalizeImage(out pos, out rot, width, height, ref intrinsics, m_PixelBuffer);
				});

				await t;

				int mapHandle = t.Result;
				int mapId = ARMap.MapHandleToId(mapHandle);
				float elapsedTime = Time.realtimeSinceStartup - startTime;

				if (mapId > 0 && ARSpace.mapIdToMap.ContainsKey(mapId))
				{
					Debug.Log(string.Format("Relocalized in {0} seconds", elapsedTime));
					stats.localizationSuccessCount++;

					ARMap map = ARSpace.mapIdToMap[mapId];
					
					if (mapId != lastLocalizedMapId)
					{
						if (resetOnMapChange)
						{
							Reset();
						}
						
						lastLocalizedMapId = mapId;
						OnMapChanged?.Invoke(mapId);
					}

					rot *= Quaternion.Euler(0f, 0f, 180.0f);
					pos = ARHelper.SwitchHandedness(pos);
					rot = ARHelper.SwitchHandedness(rot);
					
					
					MapOffset mo = ARSpace.mapIdToOffset[mapId];

					Matrix4x4 offsetNoScale = Matrix4x4.TRS(mo.position, mo.rotation, Vector3.one);
					Vector3 scaledPos = Vector3.Scale(pos, mo.scale);
					Matrix4x4 cloudSpace = offsetNoScale * Matrix4x4.TRS(scaledPos, rot, Vector3.one);
					Matrix4x4 trackerSpace = Matrix4x4.TRS(camPos, camRot, Vector3.one);
					Matrix4x4 m = trackerSpace * (cloudSpace.inverse);

					if (useFiltering)
						mo.space.filter.RefinePose(m);
					else
						ARSpace.UpdateSpace(mo.space, m.GetColumn(3), m.rotation);

					Vector3 p = m.GetColumn(3);
					Vector3 euler = m.rotation.eulerAngles;

					GetLocalizerPose(out lastLocalizedPose, mapId, pos, rot, m.inverse);
					map.NotifySuccessfulLocalization(mapId);
					OnPoseFound?.Invoke(lastLocalizedPose);
				}
				else
				{
					Debug.Log(string.Format("Localization attempt failed after {0} seconds", elapsedTime));
				}
			}
			else
			{
				Debug.LogError("No camera pixel buffer");
			}

			base.Localize();
		}

        public override async void LocalizeServer(SDKMapId[] mapIds)
        {

			if (m_PixelBuffer != IntPtr.Zero)
			{
				stats.localizationAttemptCount++;

				JobLocalizeServerAsync j = new JobLocalizeServerAsync();

				Vector4 intrinsics;
				Pose headPose = new Pose(GSXRManager.Instance.head.position, GSXRManager.Instance.head.rotation);// NRFrame.HeadPose;
				Pose rgbCameraFromEyePose = new Pose(GSXRManager.Instance.head.position, GSXRManager.Instance.head.rotation);// NRFrame.EyePoseFromHead.RGBEyePose;
				Vector3 camPos = headPose.position + rgbCameraFromEyePose.position;
				Quaternion camRot = headPose.rotation * rgbCameraFromEyePose.rotation;
				int channels = 1;
				int width = webTex.width;
				int height = webTex.height;
				byte[] pixels =  GetYBufFromTexture();

				GetIntrinsics(out intrinsics);

				float startTime = Time.realtimeSinceStartup;

				Task<(byte[], icvCaptureInfo)> t = Task.Run(() =>
				{
					byte[] capture = new byte[channels * width * height + 8192];
					icvCaptureInfo info = Immersal.Core.CaptureImage(capture, capture.Length, pixels, width, height, channels);
					Array.Resize(ref capture, info.captureSize);
					return (capture, info);
				});

				await t;

				j.image = t.Result.Item1;
				j.intrinsics = intrinsics;
				j.mapIds = mapIds;

				j.OnResult += (SDKLocalizeResult result) =>
				{
					float elapsedTime = Time.realtimeSinceStartup - startTime;

					if (result.success)
					{
						Debug.Log("*************************** On-Server Localization Succeeded ***************************");
						Debug.Log(string.Format("Relocalized in {0} seconds", elapsedTime));

						int mapId = result.map;

						if (mapId > 0 && ARSpace.mapIdToMap.ContainsKey(mapId))
						{
							ARMap map = ARSpace.mapIdToMap[mapId];

							if (mapId != lastLocalizedMapId)
							{
								if (resetOnMapChange)
								{
									Reset();
								}
								
								lastLocalizedMapId = mapId;
								OnMapChanged?.Invoke(mapId);
							}

							MapOffset mo = ARSpace.mapIdToOffset[mapId];
							stats.localizationSuccessCount++;
							
							Matrix4x4 responseMatrix = Matrix4x4.identity;
							responseMatrix.m00 = result.r00; responseMatrix.m01 = result.r01; responseMatrix.m02 = result.r02; responseMatrix.m03 = result.px;
							responseMatrix.m10 = result.r10; responseMatrix.m11 = result.r11; responseMatrix.m12 = result.r12; responseMatrix.m13 = result.py;
							responseMatrix.m20 = result.r20; responseMatrix.m21 = result.r21; responseMatrix.m22 = result.r22; responseMatrix.m23 = result.pz;

							Vector3 pos  = responseMatrix.GetColumn(3);
							Quaternion rot = responseMatrix.rotation;
							
							rot *= Quaternion.Euler(0f, 0f, 180.0f);
							pos = ARHelper.SwitchHandedness(pos);
							rot = ARHelper.SwitchHandedness(rot);
							
							Matrix4x4 offsetNoScale = Matrix4x4.TRS(mo.position, mo.rotation, Vector3.one);
							Vector3 scaledPos = Vector3.Scale(pos, mo.scale);
							Matrix4x4 cloudSpace = offsetNoScale * Matrix4x4.TRS(scaledPos, rot, Vector3.one);
							Matrix4x4 trackerSpace = Matrix4x4.TRS(camPos, camRot, Vector3.one);
							Matrix4x4 m = trackerSpace * (cloudSpace.inverse);

							if (useFiltering)
								mo.space.filter.RefinePose(m);
							else
								ARSpace.UpdateSpace(mo.space, m.GetColumn(3), m.rotation);

							double[] ecef = map.MapToEcefGet();
							LocalizerBase.GetLocalizerPose(out lastLocalizedPose, mapId, pos, rot, m.inverse, ecef);
							map.NotifySuccessfulLocalization(mapId);
							OnPoseFound?.Invoke(lastLocalizedPose);
						}
					}
					else
					{
						Debug.Log("*************************** On-Server Localization Failed ***************************");
						Debug.Log(string.Format("Localization attempt failed after {0} seconds", elapsedTime));
					}
				};

				await j.RunJobAsync();
			}
			else
			{
				Debug.LogError("No camera pixel buffer");
			}

			base.LocalizeServer(mapIds);
        }

        public override async void LocalizeGeoPose(SDKMapId[] mapIds)
        {

			if (m_PixelBuffer != IntPtr.Zero)
			{
				stats.localizationAttemptCount++;

				JobGeoPoseAsync j = new JobGeoPoseAsync();

				Vector4 intrinsics;
				Pose headPose = new Pose(GSXRManager.Instance.head.position, GSXRManager.Instance.head.rotation);// NRFrame.HeadPose;
				Pose rgbCameraFromEyePose = new Pose(GSXRManager.Instance.head.position, GSXRManager.Instance.head.rotation);// NRFrame.EyePoseFromHead.RGBEyePose;
				Vector3 camPos = headPose.position + rgbCameraFromEyePose.position;
				Quaternion camRot = headPose.rotation * rgbCameraFromEyePose.rotation;
				int channels = 1;
				int width = webTex.width;
				int height = webTex.height;
				byte[] pixels = GetYBufFromTexture();

				GetIntrinsics(out intrinsics);

				float startTime = Time.realtimeSinceStartup;

				Task<(byte[], icvCaptureInfo)> t = Task.Run(() =>
				{
					byte[] capture = new byte[channels * width * height + 8192];
					icvCaptureInfo info = Immersal.Core.CaptureImage(capture, capture.Length, pixels, width, height, channels);
					Array.Resize(ref capture, info.captureSize);
					return (capture, info);
				});

				await t;

				j.image = t.Result.Item1;
				j.intrinsics = intrinsics;
				j.mapIds = mapIds;

				j.OnResult += (SDKGeoPoseResult result) =>
				{
					float elapsedTime = Time.realtimeSinceStartup - startTime;

					if (result.success)
					{
						Debug.Log("*************************** GeoPose Localization Succeeded ***************************");
						Debug.Log(string.Format("Relocalized in {0} seconds", elapsedTime));

						int mapId = result.map;
						double latitude = result.latitude;
						double longitude = result.longitude;
						double ellipsoidHeight = result.ellipsoidHeight;
						Quaternion rot = new Quaternion(result.quaternion[1], result.quaternion[2], result.quaternion[3], result.quaternion[0]);
						Debug.Log(string.Format("GeoPose returned latitude: {0}, longitude: {1}, ellipsoidHeight: {2}, quaternion: {3}", latitude, longitude, ellipsoidHeight, rot));

						double[] ecef = new double[3];
						double[] wgs84 = new double[3] { latitude, longitude, ellipsoidHeight };
						Core.PosWgs84ToEcef(ecef, wgs84);
						if (ARSpace.mapIdToMap.ContainsKey(mapId))
						{
							ARMap map = ARSpace.mapIdToMap[mapId];
							if (mapId != lastLocalizedMapId)
							{
								if (resetOnMapChange)
								{
									Reset();
								}
								lastLocalizedMapId = mapId;
								OnMapChanged?.Invoke(mapId);
							}

							MapOffset mo = ARSpace.mapIdToOffset[mapId];
							stats.localizationSuccessCount++;
							double[] mapToEcef = map.MapToEcefGet();
							Vector3 mapPos;
							Quaternion mapRot;
							Core.PosEcefToMap(out mapPos, ecef, mapToEcef);
							Core.RotEcefToMap(out mapRot, rot, mapToEcef);
							
							Matrix4x4 offsetNoScale = Matrix4x4.TRS(mo.position, mo.rotation, Vector3.one);
							Vector3 scaledPos = Vector3.Scale(mapPos, mo.scale);
							Matrix4x4 cloudSpace = offsetNoScale * Matrix4x4.TRS(scaledPos, mapRot, Vector3.one);
							Matrix4x4 trackerSpace = Matrix4x4.TRS(camPos, camRot, Vector3.one);
							Matrix4x4 m = trackerSpace*(cloudSpace.inverse);
							
							if (useFiltering)
								mo.space.filter.RefinePose(m);
							else
								ARSpace.UpdateSpace(mo.space, m.GetColumn(3), m.rotation);
							LocalizerBase.GetLocalizerPose(out lastLocalizedPose, mapId, cloudSpace.GetColumn(3), cloudSpace.rotation, m.inverse, mapToEcef);
							map.NotifySuccessfulLocalization(mapId);
							OnPoseFound?.Invoke(lastLocalizedPose);
						}
					}
					else
					{
						Debug.Log("*************************** GeoPose Localization Failed ***************************");
						Debug.Log(string.Format("GeoPose localization attempt failed after {0} seconds", elapsedTime));
					}
				};

				await j.RunJobAsync();
			}
			else
			{
				Debug.LogError("No camera pixel buffer");
			}

			base.LocalizeGeoPose(mapIds);
		}
		public RenderTexture rtx;
		Texture2D screenShot;
		IEnumerator RenderTexturesScreenCapture()
		{
			yield return new WaitForEndOfFrame();
			if(screenShot==null)
				screenShot = new Texture2D(rtx.width, rtx.height, TextureFormat.RGB24, false);

			RenderTexture.active = rtx;
			screenShot.ReadPixels(new Rect(0, 0, rtx.width, rtx.height), 0, 0);
			screenShot.Apply();
			//Camera.main.targetTexture = null;
			RenderTexture.active = null;
			yield return 0;
			OnRGBCaptureUpdate();
		}

		private void OnRGBCaptureUpdate()
        {
			byte[] pixels = GetYBufFromTexture();

			unsafe
			{
				ulong handle;
				byte* ptr = (byte*)UnsafeUtility.PinGCArrayAndGetDataAddress(pixels, out handle);
				m_PixelBuffer = (IntPtr)ptr;
				UnsafeUtility.ReleaseGCObject(handle);
			}
        }


		private byte[] GetYBufFromTexture()
		{
			Color32[] rgbArray = screenShot.GetPixels32();
			byte[] pixels = new byte[rgbArray.Length];

			for (int i = 0; i < rgbArray.Length; i++)
			{
				pixels[i] = rgbArray[i].g;
			}

			return pixels;
		}

        private void GetIntrinsics(out Vector4 intrinsics)
        {
            intrinsics = Vector4.zero;
			
			intrinsics.x = 1153.629f;
			intrinsics.y = 1154.611f;
			intrinsics.z = 650.2759f;
			intrinsics.w = 373.7895f;
        }
	}
}
